# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# Loads .cici/config.yaml
# Defines targes, variables and metadata about the project.


import logging
import re
from pathlib import Path
from typing import Any, Optional, Union

import msgspec
import ruamel.yaml
from msgspec.structs import replace

from . import models as cici_config

decoder = msgspec.json.Decoder(type=cici_config.File)
target_decoder = msgspec.json.Decoder(type=cici_config.Target)
image_fqdn_regex = re.compile(r"^[\w\.-]+/")


# make sure each variable has its 'name' field set from its key
def inject_variable_names(variables: dict[str, Union[str, dict]]) -> dict[str, dict]:
    patched = {}
    for key, value in variables.items():
        if isinstance(value, str):
            value = {"value": value}
        # if value is None or not a dict, treat it as empty dict
        if not isinstance(value, dict):
            raise TypeError(f"Expected dict for variable {key}, got {type(value)}")
        # only add name if missing
        value.setdefault("name", key)
        patched[key] = value
    return patched


def load_targets_from_dir(target_dir: Path) -> list[cici_config.Target]:
    if not target_dir.exists() or not target_dir.is_dir():
        return []

    yaml = ruamel.yaml.YAML(typ="safe")
    targets: list[cici_config.Target] = []

    # checking for .yaml
    for file in sorted(target_dir.glob("*.yaml")):
        try:
            with open(file, "r", encoding="utf-8") as f:
                data = yaml.load(f) or {}

            # make sure name exists before decoding
            data.setdefault("name", file.stem)

            # decode using msgspec for validation and type safety
            target_obj = target_decoder.decode(msgspec.json.encode(data))
            targets.append(target_obj)

        except msgspec.ValidationError as e:
            logging.warning(f"Validation failed for {file.name}: {e}")
        except Exception as e:
            logging.warning(f"Failed to load target {file.name}: {e}")

    logging.info(f"Loaded {len(targets)} valid targets from {target_dir}")
    return targets


def resolve_targets(
    cici_config_file: Optional[cici_config.File], config_path: Path
) -> Optional[cici_config.File]:
    # Resolve and merge targets from .cici/config.yaml and from .cici/targets/*.yaml

    # - Directory targets override config.yaml targets with the same name
    # - If no targets dir exists, use the targets defined in the config.yaml
    # - returns updated 'cici_config.File' with merged targets

    if not cici_config_file:
        return None

    # get directory that could exist targets
    targets_dir = config_path.parent / "targets"

    # get targets from config.yaml
    config_targets: list[cici_config.Target] = list(cici_config_file.targets)

    # load targets from '.cici/targets/' if there are any
    dir_targets: list[cici_config.Target] = []
    if targets_dir.exists() and any(targets_dir.glob("*.yaml")):
        logging.info(f"loading targets from directory: {targets_dir}")
        dir_targets = load_targets_from_dir(targets_dir)

    # combine all targets
    all_targets = list(config_targets) + list(dir_targets)

    # get target names
    target_names = [
        target.name for target in all_targets if getattr(target, "name", None)
    ]

    # check duplicates
    if len(target_names) != len(set(target_names)):
        raise ValueError(
            f"Duplicate target names found: {', '.join(sorted(target_names))}"
        )

    # create new file with merged target list
    resolved_file = replace(cici_config_file, targets=all_targets)
    return resolved_file


def patch_image(image: str, container_proxy: str = "${CONTAINER_PROXY}") -> str:
    """Patch in $CONTAINER_PROXY to image unless the following are true:

    A: Does the image URL contain ${CONTAINER_PROXY} (the literal string)
    B: Is the image URL a fully-qualified container URL?
    C: Does the image already start with a variable?
    """

    if not image:
        return image
    if container_proxy in image:
        return image

    if image_fqdn_regex.match(image):
        return image

    if image.startswith("$"):
        return image

    return f"{container_proxy}{image}"


def loads(
    text: str,
    gitlab_ci_jobs: Optional[dict[str, Any]] = None,
    precommit_hooks: Optional[dict[str, Any]] = None,
) -> cici_config.File:
    # parse YAML into fully-typed File object
    if gitlab_ci_jobs is None:
        gitlab_ci_jobs = {}
    if precommit_hooks is None:
        precommit_hooks = {}

    yaml = ruamel.yaml.YAML(typ="safe")

    data = yaml.load(text)

    # verify targets exists even if empty
    data.setdefault("targets", [])

    # Inject precommit/gitlab include references into each target
    for target in data["targets"]:
        name = target["name"]

        # if precommit hook exists in .pre-commit-hooks.yml, attach a pre-commit hook reference
        # Injected object must match PreCommitHookTarget schema to keep msgspec happy
        if name in precommit_hooks:
            target["precommit_hook"] = {"name": name}
        else:
            # Remove key entirely when no hook exists so msgspec can apply
            target.pop("precommit_hook", None)

        target["gitlab_include"] = {"name": name}

    # Debug to test injection
    # print("After injection:", data["targets"])

    if "variables" in data:
        data["variables"] = inject_variable_names(data["variables"])

    # decode into file_struct
    file_struct = decoder.decode(msgspec.json.encode(data))

    # post process to patch CONTAINER_PROXY to container images
    patched_targets = []
    for target in file_struct.targets:
        if target.container is not None:
            patched_container = replace(
                target.container,
                image=patch_image(target.container.image),
            )
            patched_targets.append(replace(target, container=patched_container))
        else:
            patched_targets.append(target)

    return replace(file_struct, targets=patched_targets)


def load(
    file: Union[str, Path],
    gitlab_ci_jobs: Optional[dict[str, Any]] = None,
    precommit_hooks: Optional[dict[str, Any]] = None,
) -> cici_config.File:
    return loads(
        open(file).read(),
        gitlab_ci_jobs=gitlab_ci_jobs,
        precommit_hooks=precommit_hooks,
    )



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import io

import pytest
import ruamel.yaml
from ruamel.yaml.scalarstring import FoldedScalarString

from cici.providers.gitlab.yaml_style import make_scalar_string


# test the folding when handling string literals
@pytest.mark.parametrize(
    "command_text",
    [
        # long command with pipe
        "toolA subcommand with many flags and arguments that make this command exceed "
        "a typical width threshold | toolB --option value --another-option something",
        # long command with multiple pipes
        "cmd1 --foo bar --baz qux | cmd2 --long-option here | cmd3 --final-stage",
        # long command without pipe
        "someverylongcommandname with a lot of arguments and flags that keeps going "
        "and going and going until it is clearly longer than any sane line width",
        # command with && and ;
        "step1 --flag value && step2 --another value ; step3 --final",
    ],
    ids=[
        "long-command-with-pipe",
        "long-command-with-multiple-pipes",
        "long-command-no-pipe",
        "long-command-with-operators",
    ],
)
def test_styled_command_has_no_semantic_newlines(command_text):
    styled = make_scalar_string(command_text)

    # If folding is used, it must be display-only
    if isinstance(styled, FoldedScalarString):
        assert "\n" not in str(styled), (
            "FoldedScalarString must not contain real newlines; "
            "folding must be display-only"
        )


# make sure ruamel dumping never prints the blank line
@pytest.mark.parametrize(
    "command_text",
    [
        "toolA subcommand with many flags and arguments that make this command exceed "
        "a typical width threshold | toolB --option value --another-option something",
        "cmd1 --foo bar --baz qux | cmd2 --long-option here | cmd3 --final-stage",
        "someverylongcommandname with a lot of arguments and flags that keeps going "
        "and going and going until it is clearly longer than any sane line width",
    ],
    ids=[
        "yaml-dump-no-blank-line-case1",
        "yaml-dump-no-blank-line-case2",
        "yaml-dump-no-blank-line-case3",
    ],
)
def test_yaml_dump_has_no_blank_line_in_folded_scalar(command_text):
    styled = make_scalar_string(command_text)

    data = {
        "job": {
            "script": [styled, "echo ok"],
        }
    }

    yaml = ruamel.yaml.YAML(typ="rt")
    yaml.width = 120
    yaml.indent(mapping=2, sequence=4, offset=2)
    yaml.default_flow_style = False

    out = io.StringIO()
    yaml.dump(data, out)
    emitted = out.getvalue()

    lines = emitted.splitlines()

    # Find folded scalar start
    try:
        folded_start = next(
            i for i, line in enumerate(lines) if line.strip() in ("- >-", "- >")
        )
    except StopIteration:
        pytest.skip("No folded scalar emitted; nothing to assert")

    # Inspect a small window of the folded block
    folded_block = lines[folded_start : folded_start + 8]

    # Regression guard: no visually empty line inside folded scalar
    assert not any(line.strip() == "" for line in folded_block[1:]), "\n".join(
        folded_block
    )


@pytest.mark.parametrize(
    "command",
    [
        # generic echo | docker login case
        'echo "password" | docker login "registry" -u "user" --password-stdin',
        # generic long piped command
        'echo "token" | some-tool login --flag-a --flag-b --flag-c --flag-d',
        # generic curl | tar example
        "curl https://example.com/archive.tar.gz | tar -C /usr/local/bin -xzf -",
    ],
    ids=[
        "echo-pipe-docker-login",
        "echo-pipe-generic-tool",
        "curl-pipe-tar",
    ],
)
def test_folded_scalar_does_not_insert_blank_lines(command):
    """
    Folded scalars should:
    - contain no semantic newlines
    - fold visually using >-
    - not introduce blank lines in YAML output
    """

    scalar = make_scalar_string(command)

    # 1. Value must not contain semantic newlines
    assert isinstance(scalar, FoldedScalarString)
    assert "\n" not in str(scalar)

    # 2. Dump YAML
    yaml = ruamel.yaml.YAML()
    yaml.width = 120

    data = {"script": [scalar]}
    stream = io.StringIO()
    yaml.dump(data, stream)
    output = stream.getvalue()

    # 3. Must be folded
    assert ">-" in output

    # 4. Must not contain blank line inside folded block
    # (blank line would show up as double newline)
    assert "\n\n" not in output

    # 5. The command must still appear intact
    for part in command.split():
        assert part in output



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import argparse
from importlib import import_module

from ._version import __version__
from .constants import COMMANDS


def get_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--version", "-v", action="version", version=f"cici {__version__}"
    )

    subparsers = parser.add_subparsers(required=True)

    for command_name in COMMANDS:
        command = import_module(f".{command_name}", "cici.cli")
        getattr(command, f"{command_name}_parser")(subparsers=subparsers)

    return parser


def main(args=None):
    parser = get_parser()
    args = parser.parse_args(args=args)
    args.func(parser=parser, args=args)



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# Loads and serializes .gitlab-ci.yml and CI job definitions

import io
import typing
from pathlib import Path
from typing import Optional

import msgspec
import ruamel.yaml
from msgspec import ValidationError
from msgspec.structs import replace
from ruamel.yaml.scalarstring import DoubleQuotedScalarString

from cici.providers.gitlab.normalizers import (
    normalize_jobs_in_data,
    normalize_scalars,
    normalize_variables,
)

from ...config.project import models as cici_config
from . import models
from . import models as gitlab
from .utils import get_job_names
from .yaml_style import always_double_quoted, style_scalars

# decoders/encoders for msgspec
decoder = msgspec.json.Decoder(type=models.File)
encoder = msgspec.json.Encoder()


# turn plain dict into gitlab.File using msgspec
def decode_file(data: dict) -> gitlab.File:
    return decoder.decode(encoder.encode(data))


# Raumel cannot walk msgspec.Structs
# This helper is final dump step so ruamel.yaml can serialize
def to_dict(obj):
    encoded = msgspec.json.encode(obj)
    return msgspec.json.decode(encoded)


def expand_job_extends(
    jobs: dict[str, gitlab.Job],
    job: gitlab.Job,
    root_name: str,
) -> gitlab.Job:

    # If the job does not use 'extends', return it unchanged
    if not job.extends:
        return job

    # extends can be a single string or a list
    # normalize to a list so we can iterate
    extends = job.extends if isinstance(job.extends, list) else [job.extends]

    # empty dict that will accumulate parent fields
    merged: dict = {}

    # for each parent job listed in extends:
    for extend in extends:
        # recursively expand the parent first so fit the parent also has 'extends' it gets resolved too
        parent_job = expand_job_extends(jobs, jobs[extend], root_name=root_name)
        # convert expanded parent struct into plain dict to easily merge its fields
        parent_dict = to_dict(parent_job)

        # Replace merged.update with block to preserve variables from ALL parents
        for field_name, parent_value in parent_dict.items():
            existing_value = merged.get(field_name)

            if isinstance(parent_value, dict) and isinstance(existing_value, dict):
                # deep merge dict fields (ie variables) so later, parents
                # override keys without wiping earlier ones
                merged[field_name] = {**existing_value, **parent_value}
            else:
                merged[field_name] = parent_value

    # convert child job into a dict
    child_dict = to_dict(job)

    # Overlay the childs fields on top of the merged parent dict
    # but skip values that are empty so we dont wipe away necessart parent defaults
    for k, v in child_dict.items():
        if v in (None, "", [], {}, False):
            continue

        if isinstance(v, dict) and isinstance(merged.get(k), dict):
            # merge child into parent but child wins on conflicts
            merged[k] = {**merged[k], **v}
        else:
            merged[k] = v

    # remove extends key since its resolved now
    merged.pop("extends", None)

    # Rebuild fresh msgspec job struct from the merged dict
    return gitlab.Job(**merged)


def expand_jobs(file_struct: gitlab.File) -> gitlab.File:
    patched_jobs = {}
    for job_name, job in file_struct.jobs.items():
        expanded_job = expand_job_extends(file_struct.jobs, job, root_name=job_name)
        patched_jobs[job_name] = expanded_job

    return replace(file_struct, jobs=patched_jobs)


def pack_jobs(data: dict) -> dict:
    if not isinstance(data, dict):
        return data

    job_names = get_job_names(data)
    jobs = {}

    # get all job definitions
    for job_name in sorted(list(job_names)):
        job_def = data[job_name]
        if not job_def:
            continue

        # validate hidden jobs if they start with '.'
        if job_name.startswith("."):
            try:
                # validate if job is valid
                msgspec.json.Decoder(type=models.Job).decode(
                    msgspec.json.Encoder().encode(job_def)
                )
                jobs[job_name] = job_def
            except ValidationError:
                # skip hidden jobs that are not valid
                continue
        else:
            jobs[job_name] = job_def

        # remove job entry from top level
        data.pop(job_name, None)

    data["jobs"] = jobs

    # make sure variables, stages, and workflow remain top level
    return data


def unpack_jobs(data: dict) -> dict:
    if "jobs" in data:
        for name, job in list(data["jobs"].items()):
            data[name] = job
        del data["jobs"]

    return data


# make sure each variable dict includes 'name' field
def inject_variable_names(
    variables: dict[str, typing.Union[str, dict]],
) -> dict[str, dict]:
    patched = {}
    for key, value in variables.items():
        if isinstance(value, str):
            value = {"value": value}
        if not isinstance(value, dict):
            raise TypeError(f"Expected dict for variable {key}, got {type(value)}")
        value.setdefault("name", key)  # inject only in dict, not struct
        patched[key] = value

    return patched


def add_config_variables(data: dict, cici_config_file) -> dict:
    # merge variables from cici_config_file into the top-level YAML variables block
    # also maps brief -> description and `default` -> `value`.

    if not cici_config_file or not hasattr(cici_config_file, "variables"):
        return data

    config_vars = getattr(cici_config_file, "variables", {})

    if not isinstance(config_vars, dict):
        return data

    # make sure top level variable exists
    if "variables" not in data or not isinstance(data["variables"], dict):
        data["variables"] = {}

    for name, variable in config_vars.items():
        # convert msgspec.Struct -> dict
        if hasattr(variable, "__struct__") or hasattr(variable, "__annotations__"):
            variable = msgspec.to_builtins(variable)
        # verify variable info is in the dict
        if not isinstance(variable, dict):
            continue

        # map brief -> description and default -> value
        description = variable.get("brief") or variable.get("description") or ""

        value = variable.get("default") or variable.get("value") or ""

        # merged_stuff = {
        #     "description": description,
        #     "value": value,
        # }
        merged_stuff = {}
        # only keep description if it contains stuff
        if description:
            merged_stuff["description"] = description
        merged_stuff["value"] = value

        # Now merge into top level variables but don't clobber existing keys unless empty
        if name not in data["variables"]:
            data["variables"][name] = merged_stuff
        else:
            existing = data["variables"][name]
            if not isinstance(existing, dict):
                existing = {"value": str(existing)}

            existing.setdefault("description", description)
            existing.setdefault("value", value)
            data["variables"][name] = existing

    return data


def inject_container_into_job(
    file_struct: gitlab.File,
    cici_config_file: Optional[cici_config.File] = None,
) -> gitlab.File:
    # return a new file struct with containers injected into matching jobs
    if not cici_config_file:
        return file_struct

    # Build a brand new dict (copy of jobs)
    patched_jobs: dict[str, gitlab.Job] = {}

    for name, job in file_struct.jobs.items():
        # find matching target by name
        target = next(
            (target for target in cici_config_file.targets if target.name == name),
            None,
        )

        if target and target.container:
            # overwrite image with container info from config
            patched_jobs[name] = replace(
                job,
                image={
                    "name": target.container.image,
                    "entrypoint": target.container.entrypoint,
                },
            )
        else:
            # no container override then keep original job
            patched_jobs[name] = job

    # return new File struct with updated jobs
    return replace(file_struct, jobs=patched_jobs)


# Parse YAML, Normalize variables, patch jobs and decode
def loads(
    text: str, cici_config_file: typing.Optional[cici_config.File] = None
) -> gitlab.File:
    yaml = ruamel.yaml.YAML()
    data = yaml.load(text)

    # Normalize scalars
    data = normalize_scalars(data)

    # merge top-level config variables
    if cici_config_file:
        data = add_config_variables(data, cici_config_file=cici_config_file)

    # normalize jobs
    data = normalize_jobs_in_data(data)

    # normalize variables format
    data = normalize_variables(data)

    # inject variable names
    if "variables" in data:
        data["variables"] = inject_variable_names(data["variables"])

    # pack the jobs
    data = pack_jobs(data)

    # decode into msgspec struct
    file_struct = decode_file(data)

    # expand extends and anchors
    file_struct = expand_jobs(file_struct)

    # inject container defaults
    file_struct = inject_container_into_job(file_struct, cici_config_file)

    return file_struct


# convert struct back to dict, apply container injection + YAML dump
def load(
    file: typing.Union[str, Path],
    cici_config_file: typing.Optional[cici_config.File] = None,
) -> gitlab.File:
    return loads(open(file).read(), cici_config_file=cici_config_file)


def dumps(
    file: gitlab.File, cici_config_file: typing.Optional[cici_config.File] = None
) -> str:
    output = io.StringIO()
    dump(file, output, cici_config_file)
    return output.getvalue()


def dump(
    file: gitlab.File,
    stream: typing.IO,
    cici_config_file: Optional[cici_config.File] = None,
):

    if cici_config_file:
        file = inject_container_into_job(file, cici_config_file)

    data = msgspec.to_builtins(file)
    data = unpack_jobs(data)
    data = style_scalars(data)

    # DEBUG
    # job = data.get("opentofu-trivy") or {}
    # script = job.get("script") or []
    # if script:
    #     print("\n[DUMP DEBUG] opentofu-trivy script[0] type:", type(script[0]))
    #     print("[DUMP DEBUG] script[0] repr:", repr(str(script[0])))
    #     print("[DUMP DEBUG] contains \\n\\n?:", "\n\n" in str(script[0]))
    #     print("[DUMP DEBUG] contains \\n?:", "\n" in str(script[0]))
    # END DEBUG

    # user round trip mode to preserve ruamel scalar styles (FoldedScalarString etc)
    yaml = ruamel.yaml.YAML(typ="rt")
    yaml.default_flow_style = False
    yaml.explicit_start = False
    yaml.preserve_quotes = True  # respect the quotes set in style_scalars()
    yaml.indent(mapping=2, sequence=4, offset=2)
    yaml.width = 120  # prevent unwanted line wrapping
    # makes sure ruamel.yml to always emit double quoted strings """"
    yaml.representer.add_representer(DoubleQuotedScalarString, always_double_quoted)

    # blocks = [{key: value} for key, value in data.items()]
    blocks = [
        ruamel.yaml.comments.CommentedMap({key: value}) for key, value in data.items()
    ]
    fragments = []
    for block in blocks:
        text = io.StringIO()
        yaml.dump(block, text)
        # yaml.dump(data, stream)
        fragments.append(text.getvalue().rstrip())
    stream.write("\n\n".join(fragments) + "\n")



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import json
from typing import Any

import markdown
import msgspec
from jinja2 import Environment, FileSystemLoader
from ruamel.yaml import YAML

from ..config.project.serializers import load as load_cici_config
from ..constants import TEMPLATE_DIR
from ..paths import (
    get_cici_config_file_path,
    get_cici_config_gitlab_ci_file_path,
    get_cici_config_path,
    get_pre_commit_hooks_path,
    get_readme_path,
)
from ..providers.gitlab.utils import get_job_names


def to_markdown(text):
    return markdown.markdown(text)


def get_yaml_data(filename):
    yaml = YAML(typ="safe")
    return yaml.load(open(filename))


def get_gitlab_ci_jobs(gitlab_ci_file) -> dict[str, Any]:
    try:
        data = get_yaml_data(gitlab_ci_file)
    except FileNotFoundError:
        return {}
    return {job: data[job] for job in get_job_names(data) if not job.startswith(".")}


def get_precommit_hooks(precommit_hooks_file) -> dict[str, Any]:
    try:
        data = get_yaml_data(precommit_hooks_file)
    except FileNotFoundError:
        return {}
    return {hook["id"]: hook for hook in data}


def readme_command(parser, args):
    environment = Environment(
        loader=FileSystemLoader(
            [
                get_cici_config_path(),
                TEMPLATE_DIR,
            ]
        ),
    )
    environment.filters["markdown"] = to_markdown

    template = environment.get_template("README.md.j2")

    gitlab_ci_jobs = get_gitlab_ci_jobs(get_cici_config_gitlab_ci_file_path())
    precommit_hooks = get_precommit_hooks(get_pre_commit_hooks_path())

    config = load_cici_config(
        args.config_file, gitlab_ci_jobs=gitlab_ci_jobs, precommit_hooks=precommit_hooks
    )

    with open(args.output_file, "w") as handle:
        handle.write(
            template.render(
                **json.loads(msgspec.json.encode(config)),
            ).rstrip()
            + "\n"  # guarantee single line ending EOF
        )


def readme_parser(subparsers):
    parser = subparsers.add_parser("readme", help="generate pipeline readme")
    parser.add_argument(
        "-o",
        "--output",
        dest="output_file",
        default=get_readme_path(),
    )
    parser.add_argument(
        "-c",
        "--config",
        dest="config_file",
        default=get_cici_config_file_path(),
    )
    parser.set_defaults(func=readme_command)
    return parser



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import pytest

from cici.config.project.serializers import loads

CONFIG_YAML = """
name: test-pipeline
targets:
  - name: A
  - name: B
  - name: C
  - name: D
"""
# Only some targets have hooks
PRECOMMIT_HOOKS = {"A": {}, "C": {}}  # type: ignore[var-annotated]


def _target_by_name(file_obj, name: str):
    for t in file_obj.targets:
        if t.name == name:
            return t
    raise AssertionError(f"Target not found: {name}")


@pytest.mark.parametrize(
    "target_name, expected_has_hook",
    [
        ("A", True),
        ("B", False),
        ("C", True),
        ("D", False),
    ],
)
def test_precommit_hook_injection_is_name_based(
    target_name: str,
    expected_has_hook: bool,
):
    file_obj = loads(
        CONFIG_YAML,
        gitlab_ci_jobs={},  # irrelevant for this test
        precommit_hooks=PRECOMMIT_HOOKS,
    )
    target = _target_by_name(file_obj, target_name)
    # dict-or-None version
    assert (target.precommit_hook is not None) == expected_has_hook



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from enum import Enum


class WhenChoices(str, Enum):
    ALWAYS = "always"
    ACCESS = "access"
    DEVELOPMENT = "development"
    IF_NOT_PRESENT = "if-not-present"
    NEVER = "never"
    ON_SUCCESS = "on_success"
    ON_FAILURE = "on_failure"
    OTHER = "other"
    PREPARE = "prepare"
    PRODUCTION = "production"
    PULL = "pull"
    PULL_PUSH = "pull-push"
    PUSH = "push"
    STAGING = "staging"
    START = "start"
    STOP = "stop"
    TESTING = "testing"
    VERIFY = "verify"


class CachePolicies(str, Enum):
    PULL = "pull"
    PUSH = "push"
    PULL_PUSH = "pull-push"


class DeploymentTiers(str, Enum):
    PRODUCTION = "production"
    STAGING = "staging"
    TESTING = "testing"
    DEVELOPMENT = "development"
    OTHER = "other"


class EnvActions(str, Enum):
    START = "start"
    PREPARE = "prepare"
    STOP = "stop"
    VERIFY = "verify"
    ACCESS = "access"


class PullPolicies(str, Enum):
    ALWAYS = "always"
    IF_NOT_PRESENT = "if-not-present"
    NEVER = "never"


class RetryMax(int, Enum):
    ZERO = 0
    ONE = 1
    TWO = 2


class RetryWhen(str, Enum):
    ALWAYS = "always"
    UNKNOWN_FAILURE = "unknown_failure"
    SCRIPT_FAILURE = "stript_failure"
    API_FAILURE = "api_failure"
    STUCK_OR_TIMEOUT_FAILURE = "stuck_or_timeout_failure"
    RUNNER_SYSTEM_FAILURE = "runner_system_failure"
    RUNNER_UNSUPPORTED = "runner_unsupported"
    STALE_SCHEDULE = "stale_schedule"
    JOB_EXECUTION_TIMEOUT = "job_execution_timeout"
    ARCHIVED_FAILURE = "archived_failure"
    UNMET_PREREQUISITES = "umet_prerequisites"
    SCHEDULER_FAILURE = "scheduler_failure"
    DATA_INTEGRITY_FAILURE = "data_integrity_failure"


CI_FILE = ".gitlab-ci.yml"



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from typing import Annotated, Dict, List, Optional

import msgspec
from msgspec import Meta


class PreCommitHookTarget(msgspec.Struct, frozen=True, kw_only=True):
    """Custom name for a pre-commit hook target."""

    name: Annotated[str, Meta(description="Name of pre-commit hook.")]


class GitLabIncludeTarget(msgspec.Struct, frozen=True, kw_only=True):
    """Custom name for a GitLab CI/CD Include target."""

    name: Annotated[str, Meta(description="Name of GitLab include target.")]


class Group(msgspec.Struct, frozen=True, kw_only=True):
    """A logical set of pipeline targets."""

    name: Annotated[str, Meta(description="Name of this group.")]
    brief: Annotated[
        str,
        Meta(
            description="Short, one-line description of this group. Supports Markdown."
        ),
    ] = ""
    description: Annotated[
        str,
        Meta(
            description="Multi-line, long-form description of this group. Supports Markdown."
        ),
    ] = ""


class Container(msgspec.Struct, frozen=True, kw_only=True):
    """Container runtime configuration for this target."""

    image: Annotated[str, Meta(description="Container image to use for this target.")]
    entrypoint: Annotated[
        List[str], Meta(description="Container entrypoint to use for this target.")
    ] = msgspec.field(default_factory=list)


class Target(msgspec.Struct, frozen=True, kw_only=True):
    """Defines a target pipeline to be generated."""

    name: Annotated[
        str,
        Meta(
            description="Name of a pipeline target. All lower-case, hyphen-separated (`kebab-case`) is expected."
        ),
    ]
    brief: Annotated[
        str,
        Meta(
            description="Short, one-line description of a pipeline target. Supports Markdown."
        ),
    ] = ""
    description: Annotated[
        str,
        Meta(
            description="Multi-line, long-form description of a pipeline target. Supports Markdown."
        ),
    ] = ""

    groups: Annotated[
        List[str],
        Meta(
            description="Denote a logical grouping of pipeline targets.",
        ),
    ] = msgspec.field(default_factory=list)
    tags: Annotated[
        List[str],
        Meta(description="Tags for curating jobs according to their purpose."),
    ] = msgspec.field(default_factory=list)

    container: Annotated[
        Optional[Container],
        Meta(description="Container runtime configuration for this pipeline target."),
    ] = None
    precommit_hook: Annotated[
        Optional[PreCommitHookTarget],
        Meta(
            description="Configuration for the pre-commit hook.",
            extra_json_schema=dict(deprecated=True),
        ),
    ] = None
    gitlab_include: Annotated[
        Optional[GitLabIncludeTarget],
        Meta(
            description="Configuration for the GitLab CI/CD include.",
            extra_json_schema=dict(deprecated=True),
        ),
    ] = None


class VariableExample(msgspec.Struct, frozen=True, kw_only=True):
    """An example of how the provided variable should be used."""

    value: Annotated[str, Meta(description="Example value for the variable.")]
    brief: Annotated[
        str, Meta(description="Short, one-line description of a variable example.")
    ] = ""


class Variable(msgspec.Struct, frozen=True, kw_only=True):
    """Defines a variable to be consumed by the target."""

    name: Annotated[
        str, Meta(description="Name of a variable.", examples=["PYTHON_VERSION"])
    ]
    brief: Annotated[
        str,
        Meta(
            description="Short, one-line description of a variable. Supports Markdown."
        ),
    ] = ""
    default: Annotated[str, Meta(description="Default value for this variable.")] = ""
    description: Annotated[
        str,
        Meta(
            description="Multi-line, long-form description of this variable. Support Markdown."
        ),
    ] = ""
    required: Annotated[
        bool, Meta(description="Is this variable required to use the pipeline?")
    ] = False
    examples: Annotated[
        List[VariableExample],
        Meta(description="List of examples to demonstrate this variable."),
    ] = msgspec.field(default_factory=list)


class File(msgspec.Struct, frozen=True, kw_only=True):
    """Top-level cici configuration object."""

    name: Annotated[
        str,
        Meta(
            description="Name of the component. All lower-case, hyphen-separated expected.",
            examples=["crosstool-ng"],
        ),
    ]

    repo_url: Annotated[
        str,
        Meta(
            description="Web URL for the source repository for this component.",
            examples=["https://gitlab.com/saferatday0/library/cxx"],
        ),
    ] = ""

    gitlab_project_path: Annotated[
        str,
        Meta(
            description="Full project path for this component on Gitlab.",
            examples=["saferatday0/library/python"],
            extra_json_schema=dict(deprecated=True),
        ),
    ] = ""

    brief: Annotated[
        str,
        Meta(
            description="Short, one-line description of a component. Supports Markdown."
        ),
    ] = ""
    description: Annotated[
        str,
        Meta(description="Multi-line description of a component. Supports Markdown."),
    ] = ""

    groups: Annotated[List[Group], Meta(description="List of groups to declare.")] = (
        msgspec.field(default_factory=list)
    )
    targets: Annotated[
        List[Target], Meta(description="List of pipeline targets to declare.")
    ] = msgspec.field(default_factory=list)
    variables: Annotated[
        Dict[str, Variable],
        Meta(description="Dictionary of input variables to declare."),
    ] = msgspec.field(default_factory=dict)



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Union

from ruamel.yaml.scalarstring import ScalarString


# Recursivley convert ruamel ScalarStrng objects into plain str """
# So that when ruamel.yaml gives us fancy objects like LiteralScalarString('foo'),
# they will be turned into plain "foo" which msgspec can handle
# test this frst, test for each of the ifs and make sure the nesting is right.
# do simplest tests first, so one that is empty etc. Need to verify.
def normalize_scalars(obj):
    if isinstance(obj, ScalarString):
        return str(obj)
    elif isinstance(obj, dict):
        return {normalize_scalars(k): normalize_scalars(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [normalize_scalars(i) for i in obj]
    elif isinstance(obj, tuple):
        return tuple(normalize_scalars(i) for i in obj)
    else:
        return obj


# bridge between raw YAML and ruamels quirky Python objects
# and the strict Job/File models
def normalize_all(data: dict) -> dict:
    # Apply all normalizations for comparisons
    return normalize_variables(normalize_jobs_in_data(data))


def normalize_variables(data: dict, *, inside_job: bool = False) -> dict:
    # Normalize variables: keep top-level structured, flatten job-level
    # inside_job is going to tell me if i am currently in a job definition or not
    if not isinstance(data, dict):
        return data

    # Detect job-level context to know top-level or job-level
    # if the dictionary has a stage: key, it is certainly a GitLab job definition
    # if the dictionary has a script: section it too means we are inside a job
    # once we know we are inside a job, the flag stays true yay
    is_job_block = (
        "stage" in data
        or "script" in data
        or "image" in data
        or "before_script" in data
    )
    inside_job = inside_job or is_job_block

    if "variables" in data and isinstance(data["variables"], dict):
        normalized: dict[str, Union[str, dict[str, Any]]] = {}
        for key, val in data["variables"].items():

            # ----------- TOP-LEVEL VARIABLES -----------
            # inside_job = False
            if not inside_job:
                if isinstance(val, dict):
                    new_val = {}
                    # Map brief -> description
                    if "brief" in val:
                        new_val["description"] = val["brief"]
                    elif "description" in val:
                        new_val["description"] = val["description"]
                    # Map default -> value
                    if "default" in val:
                        new_val["value"] = val["default"]
                    elif "value" in val:
                        new_val["value"] = val["value"]
                    # Ensure a value key always exists
                    new_val.setdefault("value", "")
                    normalized[key] = new_val
                elif isinstance(val, str):
                    normalized[key] = {"value": val}
                else:
                    normalized[key] = {"value": str(val)}

            # ----------- JOB-LEVEL VARIABLES -----------
            # inside_job = True
            else:
                # For jobs, we always flatten to simple key:value
                if isinstance(val, dict):
                    # If it has a "value" field, unwrap it
                    if "value" in val:
                        normalized[key] = val["value"]
                    elif "default" in val:
                        normalized[key] = val["default"]
                    else:
                        # Convert dicts to a string if nothing matches
                        normalized[key] = str(val)
                else:
                    normalized[key] = val
        data["variables"] = normalized

    # ----------- Recursive traversal -----------
    # walk throuogh the entire YAML tree not just the top level dictionary
    for k, v in list(data.items()):
        if isinstance(v, dict):
            data[k] = normalize_variables(v, inside_job=inside_job)
        elif isinstance(v, list):
            data[k] = [
                (
                    normalize_variables(i, inside_job=inside_job)
                    if isinstance(i, dict)
                    else i
                )
                for i in v
            ]
    return data


def normalize_jobs_in_data(data: dict) -> dict:
    # Ensure jobs is always a dict, not a list of dicts
    if "jobs" in data and isinstance(data["jobs"], list):
        merged = {}
        for item in data["jobs"]:
            if isinstance(item, dict):
                merged.update(item)
        data["jobs"] = merged
    return data



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import hashlib
import json
import urllib.error
import urllib.parse
import urllib.request
from datetime import datetime
from io import StringIO

import ruamel.yaml
from termcolor import colored

from ..config.user import CACHE_DIR, CACHE_TIMEOUT, GITLAB_PRIVATE_TOKEN, GITLAB_URL
from ..exceptions import ComponentNotFoundError

INCLUDE_CACHE_TIMEOUT = CACHE_TIMEOUT

INCLUDE_CACHE_DIR = CACHE_DIR / "include"

API_V4_URL = f"{GITLAB_URL}/api/v4"

HEADERS: dict[str, str] = {}

if GITLAB_PRIVATE_TOKEN:
    HEADERS.update({"PRIVATE-TOKEN": GITLAB_PRIVATE_TOKEN})


def to_fragment(text):
    if not isinstance(text, str):
        return text
    return urllib.parse.quote(text, safe="")


def get_apiv4_url(url):
    try:
        url = f"{API_V4_URL}{url}"
        request = urllib.request.Request(url=url, headers=HEADERS)
        response = urllib.request.urlopen(request)
        content = response.read()
        return json.loads(content)
    except urllib.error.HTTPError:
        return None


def get_project(project_id):
    return get_apiv4_url(f"/projects/{to_fragment(project_id)}")


def get_latest_release(project_id):
    return get_apiv4_url(
        f"/projects/{to_fragment(project_id)}/releases/permalink/latest"
    )


def write_project_data(filename, project_name):
    project = get_project(project_name)
    if project is None:
        raise ComponentNotFoundError(f"component not found: {project_name}")
    release = get_latest_release(project["id"])
    with open(filename, "w") as handle:
        handle.write(
            json.dumps(
                {"project": project, "release": release},
                indent=4,
            )
        )


def update_include(include, force=False):
    if any(key not in include for key in ("project", "file")):
        return include

    includehash = hashlib.sha1(include["project"].lower().encode()).hexdigest()
    include_hash_file = INCLUDE_CACHE_DIR / f"{includehash}.json"

    if not force and include_hash_file.exists():
        current_timestamp = datetime.now().timestamp()
        file_timestamp = include_hash_file.stat().st_mtime
        if current_timestamp > file_timestamp + INCLUDE_CACHE_TIMEOUT:
            write_project_data(include_hash_file, include["project"])
    else:
        write_project_data(include_hash_file, include["project"])

    include_data = json.load(open(include_hash_file, "r"))
    project = include_data["project"]
    project_name = project["path_with_namespace"]
    latest_release = include_data["release"]

    if latest_release:
        latest_tag = latest_release["tag_name"]
        current_tag = include.get("ref", None)
        if current_tag:
            if current_tag != latest_tag:
                print(
                    colored("updated", "magenta"),
                    project_name,
                    colored("from", "magenta"),
                    current_tag,
                    colored("to", "magenta"),
                    latest_tag,
                )
        elif not current_tag:
            print(
                colored("updated", "magenta"),
                project_name,
                colored("to", "magenta"),
                latest_tag,
            )
        include["ref"] = latest_tag
    newinclude = {}
    newinclude["project"] = include["project"]
    if "ref" in include:
        newinclude["ref"] = include["ref"]
    newinclude["file"] = include["file"]
    return newinclude


def update_includes(includes, force=False):
    if not isinstance(includes, list):
        return includes

    return [update_include(include, force=force) for include in includes]


def update_command(parser, args):
    yaml = ruamel.yaml.YAML()
    data = yaml.load(open(args.filename))

    if not INCLUDE_CACHE_DIR.exists():
        INCLUDE_CACHE_DIR.mkdir(mode=0o755, parents=True, exist_ok=True)

    if "include" in data:
        try:
            data["include"] = update_includes(data["include"], force=args.force)
        except ComponentNotFoundError as excinfo:
            parser.error(excinfo)

    output_file = StringIO()

    yaml.indent(mapping=2, sequence=4, offset=2)
    yaml.dump(data, output_file)

    output = output_file.getvalue()
    output = "\n".join([line.rstrip() for line in output.splitlines()]) + "\n"

    with open(args.filename, "w") as handle:
        handle.write(output)


def update_parser(subparsers):
    parser = subparsers.add_parser(
        "update", help="pin CI includes to the latest versions"
    )
    parser.add_argument(
        "-f", "--force", action="store_true", help="update CI includes right now"
    )
    parser.add_argument("filename", nargs="?", default=".gitlab-ci.yml")
    parser.set_defaults(func=update_command)
    return parser



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from .constants import CI_FILE
from .models import File
from .serializers import dump, dumps, load, loads

__all__ = ["CI_FILE", "File", "dump", "dumps", "load", "loads"]



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from pathlib import Path
from typing import Optional

from .constants import CONFIG_DIR_NAME


def get_cici_root_path(root_path: Optional[Path] = None) -> Path:
    if root_path is None:
        root_path = Path.cwd().absolute()
    return root_path


def get_cici_config_path(root_path: Optional[Path] = None) -> Path:
    return get_cici_root_path() / CONFIG_DIR_NAME


def get_cici_config_file_path(root_path: Optional[Path] = None) -> Path:
    return get_cici_config_path() / "config.yaml"


def get_cici_config_gitlab_ci_file_path(root_path: Optional[Path] = None) -> Path:
    return get_cici_config_path() / ".gitlab-ci.yml"


def get_cici_config_readme_template_path(root_path: Optional[Path] = None) -> Path:
    return get_cici_config_path() / "README.md.j2"


def get_pre_commit_hooks_path(root_path: Optional[Path] = None) -> Path:
    return get_cici_root_path() / ".pre-commit-hooks.yaml"


def get_readme_path(root_path: Optional[Path] = None) -> Path:
    return get_cici_root_path() / "README.md"



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from pathlib import Path

BASE_DIR = Path(__file__).parent.absolute()

COMMAND_DIR = BASE_DIR / "cli"

SCHEMA_DIR = BASE_DIR / "schema"

TEMPLATE_DIR = BASE_DIR / "templates"

README_TEMPLATE_FILE = TEMPLATE_DIR / "README.md.j2"

COMMANDS = sorted(
    [path.stem for path in COMMAND_DIR.glob("*.py") if not path.stem.startswith("_")]
)

PROVIDER_DIR = BASE_DIR / "providers"

PROVIDERS = sorted(
    [path.stem for path in PROVIDER_DIR.glob("*.py") if not path.stem.startswith("_")]
    + [
        path.stem
        for path in PROVIDER_DIR.glob("*/")
        if path.is_dir() and not path.stem.startswith("_")
    ]
)

DEFAULT_PROVIDER = PROVIDERS[0]

# this really should be based on the git project directory rather than the
# working directory but this works for now
PROJECT_DIR = Path.cwd()

CONFIG_DIR_NAME = ".cici"



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Dict, List, Optional, Union

import msgspec
from msgspec import field

from .constants import (
    CachePolicies,
    DeploymentTiers,
    EnvActions,
    RetryMax,
    RetryWhen,
    WhenChoices,
)


class AllowFailure(msgspec.Struct, omit_defaults=True, frozen=True, kw_only=True):
    exit_codes: Union[int, list[int]] = field(default_factory=list)


class CoverageReport(msgspec.Struct, omit_defaults=True, frozen=True, kw_only=True):
    coverage_format: str
    path: str


class ArtifactReports(msgspec.Struct, omit_defaults=True, frozen=True, kw_only=True):
    coverage_report: Optional[CoverageReport] = None
    junit: Union[str, list[str]] = field(default_factory=list)
    terraform: str = ""
    container_scanning: str = ""


class Artifacts(msgspec.Struct, omit_defaults=True, frozen=True, kw_only=True):
    name: str = ""
    exclude: List[str] = field(default_factory=list)
    expire_in: str = ""
    expose_as: str = ""
    public: bool = True
    paths: List[str] = field(default_factory=list)
    reports: Optional[ArtifactReports] = None
    untracked: bool = False
    when: WhenChoices = WhenChoices.ON_SUCCESS


class CacheKey(msgspec.Struct, omit_defaults=True, frozen=True, kw_only=True):
    files: List[str] = field(default_factory=list)
    prefix: str = ""


class Cache(msgspec.Struct, omit_defaults=True, frozen=True, kw_only=True):
    key: Union[str, CacheKey] = ""
    paths: List[str] = field(default_factory=list)
    untracked: bool = False
    unprotect: bool = False
    when: WhenChoices = WhenChoices.ON_SUCCESS
    policy: CachePolicies = CachePolicies.PULL_PUSH
    fallback_keys: List[str] = field(default_factory=list)


class EnvironmentKubernetes(
    msgspec.Struct, omit_defaults=True, frozen=True, kw_only=True
):
    namespace: str = ""


class Environment(msgspec.Struct, omit_defaults=True, frozen=True, kw_only=True):
    name: str
    url: str = ""
    on_stop: str = ""
    action: EnvActions = EnvActions.START
    auto_stop_in: str = ""
    kubernetes: Optional[EnvironmentKubernetes] = None
    deployment_tier: Optional[DeploymentTiers] = None


class Hooks(msgspec.Struct, omit_defaults=True, frozen=True, kw_only=True):
    pre_get_sources_script: Union[str, list[Union[str, list[str]]]] = field(
        default_factory=list
    )


class IDToken(msgspec.Struct, omit_defaults=True, frozen=True, kw_only=True):
    aud: Union[str, list[str]] = field(default_factory=list)


class Image(msgspec.Struct, omit_defaults=True, frozen=True, kw_only=True, tag="image"):
    name: str
    entrypoint: Union[str, list[str]] = field(default_factory=list)
    pull_policy: Union[str, list[str]] = field(default_factory=list)


class IncludeLocal(
    msgspec.Struct, omit_defaults=True, frozen=True, kw_only=True, tag="local"
):
    local: str


class IncludeProject(
    msgspec.Struct, omit_defaults=True, frozen=True, kw_only=True, tag="project"
):
    project: str
    ref: Optional[str] = None
    file: Union[str, list[str]] = field(default_factory=list)


class IncludeRemote(
    msgspec.Struct, omit_defaults=True, frozen=True, kw_only=True, tag="remote"
):
    remote: str


class IncludeTemplate(
    msgspec.Struct, omit_defaults=True, frozen=True, kw_only=True, tag="template"
):
    template: str


class RuleChanges(msgspec.Struct, omit_defaults=True, frozen=True, kw_only=True):
    compare_to: str
    paths = List[str]


class Rule(msgspec.Struct, omit_defaults=True, frozen=True, kw_only=True):
    if_: Optional[str] = msgspec.field(default=None, name="if")
    when: str = ""
    changes: Union[List[str], RuleChanges] = field(default_factory=list)
    exists: List[str] = field(default_factory=list)
    allow_feature: bool = False
    needs: Optional[Union[List[str], Dict[str, str]]] = None
    variables: Dict[str, str] = field(default_factory=dict)


class Retry(msgspec.Struct, omit_defaults=True, frozen=True, kw_only=True):
    max: RetryMax = RetryMax.ZERO
    when: Union[RetryWhen, list[RetryWhen]] = RetryWhen.ALWAYS


class Variable(msgspec.Struct, omit_defaults=True, frozen=True, kw_only=True):
    description: Optional[str] = None
    value: Optional[str] = None
    options: Optional[list[str]] = None
    expand: bool = True


class Service(msgspec.Struct, omit_defaults=True, frozen=True, kw_only=True):
    name: str
    entrypoint: Optional[Union[str, list[str]]] = None
    command: Optional[Union[str, list[str]]] = None
    variables: Dict[str, Union[str, Variable]] = field(default_factory=dict)
    alias: Optional[str] = None
    pull_policy: Union[str, list[str]] = field(default_factory=list)


class Job(msgspec.Struct, omit_defaults=True, frozen=True, kw_only=True):
    extends: Union[str, list[str]] = []
    stage: Optional[str] = None
    image: Optional[Union[str, Image]] = None
    services: List[Union[str, Service]] = field(default_factory=list)
    variables: Dict[str, Union[str, Variable]] = field(default_factory=dict)
    before_script: Union[str, list[Union[str, list[str]]]] = field(default_factory=list)
    script: Union[str, list[Union[str, list[str]]]] = field(default_factory=list)
    after_script: Union[str, list[Union[str, list[str]]]] = field(default_factory=list)
    allow_failure: Union[bool, AllowFailure] = False
    artifacts: Optional[Artifacts] = None
    coverage: str = ""
    cache: Optional[Cache] = None
    dependencies: Optional[list] = None
    environment: Union[str, Environment] = ""
    hooks: Optional[Hooks] = None
    id_tokens: Dict[str, IDToken] = field(default_factory=dict)
    interruptible: bool = False
    needs: Optional[list[str]] = None
    retry: Union[int, Retry] = 0
    resource_group: str = ""
    rules: List[Rule] = field(default_factory=list)
    tags: List[str] = field(default_factory=list)


class Workflow(msgspec.Struct, omit_defaults=True, frozen=True, kw_only=True):
    name: Optional[str] = None
    rules: List[Rule] = field(default_factory=list)


class Default(msgspec.Struct, omit_defaults=True, frozen=True, kw_only=True):
    after_script: Union[str, list[Union[str, list[str]]]] = field(default_factory=list)
    artifacts: Optional[Artifacts] = None
    before_script: Union[str, list[Union[str, list[str]]]] = field(default_factory=list)
    cache: Optional[Cache] = None
    hooks: Optional[Hooks] = None
    image: Optional[Union[str, Image]] = None
    interruptible: bool = False
    retry: Union[int, Retry] = 0
    tags: List[str] = field(default_factory=list)


class File(msgspec.Struct, omit_defaults=True, frozen=True, kw_only=True):
    jobs: Dict[str, Job] = field(default_factory=dict)
    stages: List[str] = field(default_factory=list)
    include: Union[
        str, List[Union[IncludeLocal, IncludeProject, IncludeRemote, IncludeTemplate]]
    ] = msgspec.field(default_factory=list)
    workflow: Optional[Workflow] = None
    variables: Dict[str, Union[str, Variable]] = field(default_factory=dict)
    extras: Dict[str, Any] = field(default_factory=dict)



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Optional

from ruamel.yaml.scalarstring import (
    DoubleQuotedScalarString,
    FoldedScalarString,
    LiteralScalarString,
    PreservedScalarString,
)


# to force ruamel.yml to always emit double quoted strings """" and not single "''"
def always_double_quoted(dumper, data):
    return dumper.represent_scalar("tag:yaml.org,2002:str", data, style='"')


def wrap_if_long(s: str, width: int = 120) -> str:
    # Wrap long strings into multiple lines for YAML folding.
    if len(s) <= width:
        return s
    # break on spaces without splitting words
    parts = []
    while len(s) > width:
        split_at = s.rfind(" ", 0, width)
        if split_at == -1:
            split_at = width
        parts.append(s[:split_at].rstrip())
        s = s[split_at:].lstrip()
    parts.append(s)
    return "\n".join(parts)


# handling string literals
def make_scalar_string(line: str, quote: bool = False):
    # Return an appropriate YAML scalar string based on content.
    if not isinstance(line, str):
        return line

    unindented = line.lstrip()

    # Preserve explicit multiline blocks
    if "\n" in unindented:
        return PreservedScalarString(line)

    # fix extra - >- between lines accidentally created with echo commands
    if unindented.startswith("echo "):
        stripped = unindented.strip()

        if len(stripped) < 100 and not any(
            sym in stripped for sym in ("&&", ";", "\\", "|")
        ):
            return stripped

        folded_scalar = FoldedScalarString(stripped)

        # If its piping "echo... | docker login ...", fold at the pipe for correct display
        if " | " in stripped:
            pipe_position = stripped.index(" | ")
            folded_scalar.fold_pos = [pipe_position]  # type: ignore[attr-defined]

        return folded_scalar

    # Folding long lines logic (like a script that is super long so that it does not do anything weird in the folding process)
    if unindented.startswith(("docker ", "helm ", "tar ", "curl ")):
        command_string = unindented.strip()

        folded_scalar = FoldedScalarString(command_string)

        if " | " in command_string:
            pipe_position = command_string.index(" | ")
            # mypy does not like fold_pos however it solves an issue that breaks the .gitlab-ci.yaml
            folded_scalar.fold_pos = [pipe_position]  # type: ignore[attr-defined]

        return folded_scalar

    # Multi-command sequences get folded
    if any(sym in unindented for sym in ("&&", ";", "\\")):
        return FoldedScalarString(wrap_if_long(unindented))

    # Long assignments get folded
    if (
        "=" in unindented
        and not unindented.startswith(("export ", "set ", "$"))
        and len(unindented) > 100
    ):
        return FoldedScalarString(wrap_if_long(unindented))

    # Explicit quoting when requested
    if quote:
        return DoubleQuotedScalarString(unindented)

    # Default: return plain string
    return unindented


# Rule engine for how YAML should look when written back out
# recursively walk through a Python structure and apply the rules defined
def style_scalars(
    obj: Any, *, quote_keys: Optional[set[str]] = None, parent_key: Optional[str] = None
) -> Any:
    # Default: keys that often need quoting in GitLab CI

    # DEBUGGING:
    # if parent_key == "opentofu-production-trivy":
    #     print("*****DEBUG****** job keys:", list(obj.keys()))

    if quote_keys is None:
        quote_keys = {"variables", "image", "services", "script", "before_script"}

    # if it is a dict, process each key/value recursively
    if isinstance(obj, dict):
        styled: dict[str, Any] = {}

        for k, v in list(obj.items()):
            v = obj[k]
            # 1. keep GitLab workflow 'if' rules unquoted
            if k == "if" and isinstance(v, str):
                styled[k] = v

            # 2. for image, only quote if contains ":" or starts with '$' (i.e. tag or env)
            elif k == "image" and isinstance(v, str):
                # removes any leading or trailing whitespace
                stripped = v.strip()
                if (
                    ":" in stripped
                    or stripped.startswith("$")
                    or stripped.startswith("${")
                ):
                    styled[k] = DoubleQuotedScalarString(v)
                else:
                    styled[k] = v

            # 3. ensures service names like
            # ${CONTAINER_PROXY}docker:28-dind
            # becomes
            # "${CONTAINER_PROXY}docker:28-dind"
            elif k == "services" and isinstance(v, list):
                styled_services = []

                for service in v:
                    if isinstance(service, dict):
                        styled_service = {}
                        for service_key, service_val in service.items():
                            if isinstance(service_val, str):
                                stripped = service_val.strip()
                                if (
                                    stripped.startswith("$")
                                    or stripped.startswith("${")
                                    or ":" in stripped
                                ):
                                    styled_service[service_key] = (
                                        DoubleQuotedScalarString(stripped)
                                    )
                                else:
                                    styled_service[service_key] = stripped  # type: ignore[assignment]
                            else:
                                styled_service[service_key] = style_scalars(
                                    service_val, quote_keys=quote_keys, parent_key=k
                                )
                        styled_services.append(styled_service)
                    else:
                        styled_services.append(
                            style_scalars(service, quote_keys=quote_keys)
                        )

                styled[k] = styled_services

            # 4. always emit description as '|-' block style
            elif k == "description" and isinstance(v, str):
                styled[k] = PreservedScalarString(v)

            # 5. for value, only quote if it looks like plain text, not env vars
            elif k == "value" and isinstance(v, str):
                stripped = v.strip()
                # (NEW) explicitly double-quote empty strings so ruamel emits ""
                if stripped == "":
                    styled[k] = DoubleQuotedScalarString("")
                # always double-quote env variables or any string starting with $
                elif stripped.startswith("$") or stripped.startswith("${"):
                    styled[k] = DoubleQuotedScalarString(stripped)
                # quote if there are dots or dashes
                elif "." in stripped or "-" in stripped:
                    styled[k] = DoubleQuotedScalarString(stripped)
                # otherwise leave it plain
                else:
                    styled[k] = stripped

            # 6. if we are INSIDE variables block, double quote all the strings
            elif k == "variables" and isinstance(v, dict):
                styled_vars: dict[str, Any] = {}

                for var_name, var_val in v.items():
                    # need to flatten the dicts into simple strings so:

                    # variable:
                    #   VARIABLE_HERE:
                    #     value: "$VARIABLE_VALUE"

                    # needs to become

                    # variable:
                    #   VARIABLE_HERE: "$VARIABLE_VALUE"
                    if (
                        isinstance(var_val, dict)
                        and "value" in var_val
                        and len(var_val) == 1
                    ):
                        var_val = var_val["value"]

                    if isinstance(var_val, str):
                        stripped = var_val.strip()

                        # quote env vars ($VAR)
                        if stripped.startswith("$") or stripped.startswith("${"):
                            styled_vars[var_name] = DoubleQuotedScalarString(stripped)

                        # quote ".", "-", or anything with "." in it
                        # also where dealing with /certs to become "/certs"
                        elif stripped in {"."} or any(
                            c in stripped for c in [".", "-", "_", "/"]
                        ):
                            styled_vars[var_name] = DoubleQuotedScalarString(stripped)

                        # quote numeric-looking values for example:

                        # variables:
                        #   GIT_DEPTH: '1'

                        # becomes

                        # variables:
                        #   GIT_DEPTH: "1"
                        elif (
                            stripped.isdigit() or stripped.replace(".", "", 1).isdigit()
                        ):
                            styled_vars[var_name] = DoubleQuotedScalarString(stripped)

                        # leave everything else plain
                        else:
                            styled_vars[var_name] = stripped

                    else:
                        styled_vars[var_name] = style_scalars(
                            var_val, quote_keys=quote_keys, parent_key=parent_key
                        )
                styled[k] = styled_vars

            # 7. in the environment block, preserve both name and action

            #   environment:
            #       name: $OPENTOFU_STATE_NAME/apply
            #       action: stop
            elif k == "environment" and isinstance(v, dict):
                styled_env = {}
                for env_key, env_val in v.items():
                    if isinstance(env_val, str):
                        styled_env[env_key] = DoubleQuotedScalarString(env_val)
                    else:
                        styled_env[env_key] = style_scalars(
                            env_val, quote_keys=quote_keys, parent_key=k
                        )
                styled[k] = styled_env

            # 8. quote values inside of id_tokens so things like sigstore in id_tokens becomes "sigstore"
            elif k == "id_tokens" and isinstance(v, dict):
                styled_tokens = {}
                for token_name, token_val in v.items():
                    if isinstance(token_val, dict):
                        styled_tokens[token_name] = {
                            tk: (
                                DoubleQuotedScalarString(tv)
                                if isinstance(tv, str)
                                else tv
                            )
                            for tk, tv in token_val.items()
                        }
                    else:
                        styled_tokens[token_name] = token_val
                styled[k] = styled_tokens

            # 9. everything else, recurse normally
            else:
                styled[k] = style_scalars(v, quote_keys=quote_keys, parent_key=k)

        # reinsert any keys skipped during iteration ie dependencies
        for k, v in obj.items():
            if k not in styled:
                styled[k] = v

        return styled

    # flatten nested lists, preserve literal strings (>-)
    elif isinstance(obj, list):
        flat: list[Any] = []
        for i in obj:
            flat.extend(i if isinstance(i, list) else [i])

        styled_list = []
        for item in flat:
            if isinstance(item, str):
                # multi-line shell to literal block
                if "\n" in item:
                    styled_list.append(LiteralScalarString(item))

                # if we are in a shell-life section, format line nicely
                elif parent_key in {"script", "before_script", "after_script"}:
                    stripped_item = item.strip()

                    # keep raw shell assignements in plain text (avoid escaping inside quotes)
                    if "=" in stripped_item.split()[0] and not stripped_item.startswith(
                        "$"
                    ):
                        styled_list.append(stripped_item)  # type: ignore[arg-type]
                    else:
                        styled_list.append(
                            make_scalar_string(
                                item if len(item) > 10 else stripped_item
                            )
                        )

                else:
                    styled_list.append(
                        style_scalars(
                            item, quote_keys=quote_keys, parent_key=parent_key
                        )
                    )
            else:
                styled_list.append(
                    style_scalars(item, quote_keys=quote_keys, parent_key=parent_key)
                )
        return styled_list or obj

    # handle plain strings
    elif isinstance(obj, str):
        stripped = obj.strip()

        # 1. don't quote variables or expressions
        if stripped.startswith("$") or stripped.startswith("${"):
            return stripped

        # 2. no quotes for "."
        if stripped == ".":
            return stripped

        # 3. don't quote well-known CI/CD keywords or literal identifiers
        UNQUOTED_KEYWORDS = {
            "test",
            "build",
            "deploy",
            "never",
            "always",
            "true",
            "false",
            "on_success",
            "on_failure",
            "manual",
            "cobertura",  # coverage format
            "sigstore",
        }
        if stripped in UNQUOTED_KEYWORDS:
            return stripped

        # 4
        if "/" in stripped or "." in stripped or "-" in stripped or "_" in stripped:
            return DoubleQuotedScalarString(stripped)

        # 5. everything else stays plain
        return stripped

    return obj



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import pytest
import ruamel.yaml

from cici.providers.gitlab.normalizers import (
    normalize_jobs_in_data,
    normalize_scalars,
    normalize_variables,
)
from cici.providers.gitlab.serializers import add_config_variables

#
#   Tests for normalize_scalar
#


@pytest.mark.parametrize(
    "input_yaml, output_yaml",
    [
        (
            "{}",
            {},
        ),
    ],
)
def test_normalize_scalar_ruamel(input_yaml, output_yaml):
    yaml = ruamel.yaml.YAML()
    ruamel_data = yaml.load(input_yaml)
    result = normalize_scalars(ruamel_data)
    assert output_yaml == result


@pytest.mark.parametrize(
    "input_yaml, output_yaml",
    [
        (
            '"test-key": "test-value"',
            {"test-key": "test-value"},
        ),
    ],
)
def test_normalize_scalar_dict(input_yaml, output_yaml):
    yaml = ruamel.yaml.YAML()
    ruamel_data = yaml.load(input_yaml)
    result = normalize_scalars(ruamel_data)
    assert output_yaml == result


@pytest.mark.parametrize(
    "input_yaml, output_yaml",
    [
        (
            'outer-test: {"inner-test-key": "inner-test-value"}',
            {"outer-test": {"inner-test-key": "inner-test-value"}},
        ),
    ],
)
def test_normalize_scalar_nested_dict(input_yaml, output_yaml):
    yaml = ruamel.yaml.YAML()
    ruamel_data = yaml.load(input_yaml)
    result = normalize_scalars(ruamel_data)
    assert output_yaml == result


@pytest.mark.parametrize(
    "input_yaml, output_yaml",
    [
        (
            """
            - one
            - two
            - 3
            """,
            ["one", "two", 3],
        ),
    ],
)
def test_normalize_scalar_list(input_yaml, output_yaml):
    yaml = ruamel.yaml.YAML()
    ruamel_data = yaml.load(input_yaml)
    result = normalize_scalars(ruamel_data)
    assert output_yaml == result


@pytest.mark.parametrize(
    "input_yaml, output_yaml",
    [
        (
            """
            -
              - a
              - b
            - c
            """,
            [["a", "b"], "c"],
        ),
    ],
)
def test_normalize_scalar_nested_list(input_yaml, output_yaml):
    yaml = ruamel.yaml.YAML()
    ruamel_data = yaml.load(input_yaml)
    result = normalize_scalars(ruamel_data)
    assert output_yaml == result


# ruamel.yaml's safe loader never returns tuples thus testing w/o ruamel YAML
@pytest.mark.parametrize(
    "input_obj, output",
    [
        # simple flat tuples
        ((1, 2, 3), (1, 2, 3)),
        (("a", "b", "c"), ("a", "b", "c")),
        # nested tuples
        (((1, 2), (3, 4)), ((1, 2), (3, 4))),
        ((("a", "b"), ("c", "d")), (("a", "b"), ("c", "d"))),
        # deeply nested tuple
        ((1, (2, (3, (4,)))), (1, (2, (3, (4,))))),
    ],
)
def test_normalize_scalar_tuple(input_obj, output):
    assert normalize_scalars(input_obj) == output


@pytest.mark.parametrize(
    "input_yaml, output",
    [
        (
            # ruamel DoubleQuotedScalarString object -> Python plain str
            'message: "test message"',
            {"message": "test message"},
        ),
        (
            # ruamel SingleQuotedScalarString -> Python plain str
            "message: 'test message'",
            {"message": "test message"},
        ),
        (
            # fancy ruamel objects -> normal str key
            '"testing message": hello!',
            {"testing message": "hello!"},
        ),
        (
            # ruamel FoldedScalarString (>) -> plain str
            "test message: >\n line1\n line2\n",
            {"test message": "line1 line2\n"},
        ),
        (
            # ruamel FoldedScalarString (>) nested -> plain str
            "items:\n  - >\n    line1\n    line2\n  - value2",
            {"items": ["line1 line2\n", "value2"]},
        ),
        (
            # ruamel LiteralScalarString (|) -> plain str with preserved newlines
            "test message: |\n line1\n line2\n",
            {"test message": "line1\nline2\n"},
        ),
        (
            # ruamel LiteralScalarString (|) nested -> plain str with preserved newlines
            "items:\n  - >\n    line1\n    line2\n  - value2",
            {"items": ["line1 line2\n", "value2"]},
        ),
        (
            # nested single quoted string in a list -> Python plain str
            "items:\n"
            "  - 'first'\n"
            "  - 'second'\n"
            "  - nested:\n"
            "      - 'third'\n",
            {"items": ["first", "second", {"nested": ["third"]}]},
        ),
    ],
)
def test_normalize_scalar_scalarstring(input_yaml, output):
    yaml = ruamel.yaml.YAML(typ="rt")
    ruamel_data = yaml.load(input_yaml)
    result = normalize_scalars(ruamel_data)
    assert output == result


@pytest.mark.parametrize(
    "input_obj, output",
    [
        (
            # Stages untouched
            {"stages": ["build", "test"]},
            {"stages": ["build", "test"]},
        ),
        (
            # empty jobs list becomes dict
            {"jobs": []},
            {"jobs": {}},
        ),
        (
            # jobs as list of dicts become merged dict
            {
                "jobs": [
                    {"job1": {"script": "echo hello"}},
                    {"job2": {"script": "echo later"}},
                ]
            },
            {
                "jobs": {
                    "job1": {"script": "echo hello"},
                    "job2": {"script": "echo later"},
                }
            },
        ),
        (
            # jobs already a dict remain unchanged
            {"jobs": {"job1": {"script": "echo hello"}}},
            {"jobs": {"job1": {"script": "echo hello"}}},
        ),
    ],
)
def test_normalize_jobs_in_data(input_obj, output):
    assert normalize_jobs_in_data(input_obj) == output


@pytest.mark.parametrize(
    "input_yaml, expected",
    [
        # plain string var normalized to {"value": "..."}
        (
            """
            variables:
              SIMPLE_VAR: "hello"
            """,
            {"variables": {"SIMPLE_VAR": {"value": "hello"}}},
        ),
        # dict with value + description preserved as-is
        (
            """
            variables:
              MY_VAR:
                description: "Some description"
                value: "123"
            """,
            {
                "variables": {
                    "MY_VAR": {"description": "Some description", "value": "123"}
                }
            },
        ),
        # dict with only description should auto-add empty value
        (
            """
            variables:
              BAD_VAR:
                description: "Missing value"
            """,
            {"variables": {"BAD_VAR": {"description": "Missing value", "value": ""}}},
        ),
        # completely empty var should not become {}
        (
            """
            variables:
              EMPTY_VAR: {}
            """,
            {"variables": {"EMPTY_VAR": {"value": ""}}},
        ),
    ],
)
def test_normalize_variables_cases(input_yaml, expected):
    yaml = ruamel.yaml.YAML(typ="safe")
    ruamel_data = yaml.load(input_yaml)
    result = normalize_variables(ruamel_data)
    assert result == expected


@pytest.mark.parametrize(
    "config_vars, expected",
    [
        (
            {
                "OPENTOFU_MODULE_VERSION": {
                    "brief": "Opentofu module version.",
                    "default": "${CI_COMMIT_TAG}",
                },
                "OPENTOFU_MODULE_NAME": {
                    "brief": "Opentofu module project name.",
                    "default": "${CI_PROJECT_NAME}",
                },
                "OPENTOFU_MODULE_DIR": {
                    "brief": "Opentofu module project directory.",
                    "default": "${CI_PROJECT_DIR}",
                },
            },
            # Expected normalized output
            {
                "OPENTOFU_MODULE_VERSION": {
                    "description": "Opentofu module version.",
                    "value": "${CI_COMMIT_TAG}",
                },
                "OPENTOFU_MODULE_NAME": {
                    "description": "Opentofu module project name.",
                    "value": "${CI_PROJECT_NAME}",
                },
                "OPENTOFU_MODULE_DIR": {
                    "description": "Opentofu module project directory.",
                    "value": "${CI_PROJECT_DIR}",
                },
            },
        ),
    ],
)
def test_add_and_normalize_variables_from_config(config_vars, expected):
    data = {"stages": ["test"], "variables": {}}

    # make a cici_config_file like object with variables in it
    class MockConfig:
        def __init__(self, vars_dict):
            self.variables = vars_dict

    mock_conif = MockConfig(config_vars)

    # do the part when config merged with vars and then normalize
    merged_stuff = add_config_variables(data, mock_conif)
    normalized = normalize_variables(merged_stuff)

    # top level variables should have description + value and not be {} (empty)
    for k, v in expected.items():
        assert k in normalized["variables"], f"{k} missing in nomralized output"
        assert "description" in normalized["variables"][k]
        assert "value" in normalized["variables"][k]
        assert normalized["variables"][k] == v



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import pytest

from cici.providers.gitlab import models as gitlab
from cici.providers.gitlab.serializers import expand_job_extends


@pytest.mark.parametrize(
    "parents, child, expected",
    [
        # multiple parents merge variables without loss
        (
            {
                "parent_a": {"A": "1", "B": "1"},
                "parent_b": {"C": "1"},
            },
            {
                "extends": ["parent_a", "parent_b"],
                "variables": {},
            },
            {
                "A": "1",
                "B": "1",
                "C": "1",
            },
        ),
        # child overrides parent but keeps other keys
        (
            {
                "parent_a": {"A": "1", "B": "1"},
            },
            {
                "extends": ["parent_a"],
                "variables": {"B": "2"},
            },
            {
                "A": "1",
                "B": "2",
            },
        ),
        # later parent overrides earlier parent on conflicts
        (
            {
                "parent_a": {"A": "1"},
                "parent_b": {"A": "2"},
            },
            {
                "extends": ["parent_a", "parent_b"],
                "variables": {},
            },
            {
                "A": "2",
            },
        ),
        # later parent overrides earlier parent on conflicts but keeps previous values
        (
            {
                "parent_a": {"A": "1", "B": "1"},
                "parent_b": {"A": "2", "C": "3"},
            },
            {
                "extends": ["parent_a", "parent_b"],
                "variables": {},
            },
            {
                "A": "2",
                "B": "1",
                "C": "3",
            },
        ),
    ],
    ids=[
        "parents merge without loss",
        "child overrides parent",
        "later parent overrides earlier parent",
        "merge and override",
    ],
)
def test_expand_job_extends_variable_merging(parents, child, expected):
    jobs = {}
    # create parent jobs
    for name, variables in parents.items():
        jobs[name] = gitlab.Job(variables=variables)
    # create child job
    jobs["child"] = gitlab.Job(
        extends=child["extends"],
        variables=child["variables"],
    )
    expanded = expand_job_extends(
        jobs,
        jobs["child"],
        root_name="child",
    )
    assert expanded.extends in (None, [], "")
    assert expanded.variables == expected



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import difflib
import filecmp
import os
import shutil
from contextlib import contextmanager
from pathlib import Path

import pytest

from cici.constants import BASE_DIR
from cici.main import main as cici

FIXTURES_DIR = BASE_DIR / ".." / "tests" / "fixtures"


@contextmanager
def pushd(path):
    oldpwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(oldpwd)


@pytest.mark.parametrize(
    "platform,name",
    [
        (
            "gitlab",
            "helm",
        ),
        (
            "gitlab",
            "terraform",
        ),
        (
            "gitlab",
            "reports",
        ),
        (
            "gitlab",
            "list-anchors",
        ),
        (
            "gitlab",
            "service-key",
        ),
        (
            "gitlab",
            "job-variables",
        ),
    ],
)
def test_end_to_end_bundle(platform, name, tmp_path):
    fixture_dir = FIXTURES_DIR / platform / name
    test_dir = tmp_path
    # test_dir = fixture_dir

    test_cici_dir = test_dir / ".cici"
    test_cici_dir.mkdir()

    files = [".cici/.gitlab-ci.yml", *[path.name for path in fixture_dir.glob("*.yml")]]
    cici_config_file = fixture_dir / ".cici" / "config.yaml"
    if cici_config_file.exists():
        files.append(".cici/config.yaml")
    for file in files:
        shutil.copyfile(fixture_dir / file, test_dir / file)
    with pushd(test_dir):
        print([path.name for path in Path.cwd().glob("*")])
        cici(["bundle"])

    for file in files:
        diff = list(
            difflib.unified_diff(
                open(fixture_dir / file).read().splitlines(keepends=True),
                open(test_dir / file).read().splitlines(keepends=True),
                fromfile=f"{file}, expected",
                tofile=f"{file}, actual",
            )
        )
        if diff:
            raise ValueError("outputs are different:\n{}".format("".join(diff)))


@pytest.mark.parametrize(
    "platform,name",
    [
        (
            "gitlab",
            "library-validator",
        ),
    ],
)
def test_end_to_end_readme(platform, name, tmp_path):
    fixture_dir = FIXTURES_DIR / platform / name
    test_dir = tmp_path

    test_cici_dir = test_dir / ".cici"
    test_cici_dir.mkdir()

    files = [".cici/README.md.j2", ".cici/config.yaml", "README.md"]
    for file in files:
        shutil.copyfile(fixture_dir / file, test_dir / file)
    with pushd(test_dir):
        print([path.name for path in Path.cwd().glob("*")])
        cici(["readme"])

    match, mismatch, errors = filecmp.cmpfiles(
        fixture_dir, test_dir, files, shallow=True
    )
    print(match, mismatch, errors)



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import textwrap
from typing import Optional

import pytest

from cici.config.project.serializers import load, resolve_targets


def strip_proxy(image: Optional[str]) -> Optional[str]:
    if not image:
        return image
    if image and image.startswith("${CONTAINER_PROXY}"):
        return image[len("${CONTAINER_PROXY}") :]
    return image


@pytest.mark.parametrize(
    "config_yaml, target_files, expected_names, expected_images",
    [
        # inline-only targets in config.yaml
        (
            """
            name: blah
            targets:
              - name: alpha
                container:
                  image: ubuntu:22.04
            """,
            {},  # no targets directory
            ["alpha"],
            ["ubuntu:22.04"],
        ),
        # both inline + directory (no duplicate names)
        (
            """
            name: blah
            targets:
              - name: theta
                container:
                  image: old:version
              - name: delta
                container:
                  image: busybox
            """,
            {
                "gamma.yaml": """
                name: gamma
                container:
                  image: alpine:latest
                """
            },
            ["theta", "delta", "gamma"],
            ["old:version", "busybox", "alpine:latest"],
        ),
    ],
)
# resolve_targets correctly handles inline, directory, and mixed
def test_resolving_targets(
    tmp_path, config_yaml, target_files, expected_names, expected_images
):
    # setup test directory structure
    cici_dir = tmp_path / ".cici"
    cici_dir.mkdir()

    config_path = cici_dir / "config.yaml"
    config_path.write_text(textwrap.dedent(config_yaml))

    targets_dir = cici_dir / "targets"
    if target_files:
        targets_dir.mkdir()
        for filename, content in target_files.items():
            (targets_dir / filename).write_text(textwrap.dedent(content))

    # load config and resolve
    config_file = load(config_path)
    resolved = resolve_targets(config_file, config_path)

    # assertions
    names = [target.name for target in resolved.targets]
    images = [
        getattr(target.container, "image", None)
        for target in resolved.targets
        if getattr(target, "container", None)
    ]

    normalized_images = [strip_proxy(img) for img in images]

    assert sorted(names) == sorted(expected_names)
    assert sorted(normalized_images) == sorted(expected_images)


@pytest.mark.parametrize(
    "image, expected",
    [
        # Image without registry should be prefixed
        ("ubuntu:22.04", "${CONTAINER_PROXY}ubuntu:22.04"),
        # Fully qualified image should remain unchanged
        ("docker.io/library/python:3.11", "docker.io/library/python:3.11"),
        # Already proxied image should remain unchanged
        (
            "${CONTAINER_PROXY}custom/image:latest",
            "${CONTAINER_PROXY}custom/image:latest",
        ),
        # Starts with variable (e.g. GitLab CI variable) should remain unchanged
        ("$CI_REGISTRY/myimage:latest", "$CI_REGISTRY/myimage:latest"),
    ],
)
# test that patch_image correctly applies or skips ${CONTAINER_PROXY} prefix.
def test_container_proxy_injection(tmp_path, image, expected):
    config_yaml = f"""
    name: test-component
    targets:
      - name: foo
        container:
          image: {image}
    """
    cici_dir = tmp_path / ".cici"
    cici_dir.mkdir()
    config_path = cici_dir / "config.yaml"
    config_path.write_text(textwrap.dedent(config_yaml))
    config_file = load(config_path)
    resolved = resolve_targets(config_file, config_path)
    assert len(resolved.targets) == 1
    actual_image = resolved.targets[0].container.image
    assert actual_image == expected


@pytest.mark.parametrize(
    "config_yaml, target_files, expected_error",
    [
        # Duplicate accross config.yaml and targets dir
        (
            """
            name: test-duplicate
            targets:
              - name: duplicate
                container:
                  image: old:version
            """,
            {
                "duplicate.yaml": """
                name: duplicate
                container:
                  image: alpine:latest
                """,
            },
            "Duplicate target names found",
        ),
        # Duplicates in config.yaml
        (
            """
            name: test-duplicate
            targets:
              - name: whatever
                container:
                  image: ubuntu:22.04
              - name: whatever
                container:
                    image: busybox
            """,
            {},
            "Duplicate target names found",
        ),
        # Duplicates in the targets dir
        (
            """
            name: test-duplicate
            targets: []
            """,
            {
                "one.yaml": """
                name: repeat
                container:
                    image: alpine:3.18
                """,
                "two.yaml": """
                name: repeat
                container:
                    image: alpine:3.19
                """,
            },
            "Duplicate target names found",
        ),
    ],
)
# test that duplicate targets raise ValueError:
def test_resolve_targets_fails_with_duplicate_names(
    tmp_path, config_yaml, target_files, expected_error
):
    cici_dir = tmp_path / ".cici"
    cici_dir.mkdir()

    config_path = cici_dir / "config.yaml"
    config_path.write_text(textwrap.dedent(config_yaml))

    targets_dir = cici_dir / "targets"

    if target_files:
        targets_dir.mkdir()
        for filename, content in target_files.items():
            (targets_dir / filename).write_text(textwrap.dedent(content))

    config_file = load(config_path)

    with pytest.raises(ValueError, match=expected_error):
        resolve_targets(config_file, config_path)



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import functools
import json
import typing

import jsonschema

from ...constants import SCHEMA_DIR


@functools.cache
def load_schema() -> typing.Any:
    schema_file = SCHEMA_DIR / "gitlab-ci.json"
    return json.load(open(schema_file))


def validate(data: typing.Any):
    schema = load_schema()
    jsonschema.validate(data, schema=schema)


def get_reserved_words():
    return {key for key in load_schema()["properties"].keys() if key not in ("pages",)}


def get_job_names(data):
    return {key for key in data.keys() if key not in get_reserved_words()}



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import os
from pathlib import Path

from appdirs import user_cache_dir, user_config_dir  # type: ignore
from decouple import Config, RepositoryIni  # type: ignore

CONFIG_FILE = Path(
    os.environ.get(
        "CICI_CONFIG_FILE", Path(user_config_dir("cici-tools")) / "config.ini"
    )
)

if not CONFIG_FILE.exists():
    CONFIG_FILE.parent.mkdir(exist_ok=True, parents=True)
    CONFIG_FILE.touch()

_config = Config(RepositoryIni(CONFIG_FILE))

CACHE_DIR = _config("CICI_CACHE_DIR", cast=Path, default=user_cache_dir("cici-tools"))

CACHE_TIMEOUT = _config("CICI_CACHE_TIMEOUT", cast=int, default=43200)

GITLAB_URL = _config("CICI_GITLAB_URL", default="https://gitlab.com")

GITLAB_PRIVATE_TOKEN = _config("CICI_GITLAB_PRIVATE_TOKEN", default="")



#!/usr/bin/env python3
# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import argparse
import json
import sys

import markdown
from jinja2 import Environment


def to_markdown(text):
    return markdown.markdown(text)


def get_ref_name(ref):
    return ref.split("/")[-1]


def get_ref_struct(schema, ref):
    return schema["$defs"][get_ref_name(ref)]


def render_header(level, text):
    return "{} {}".format("#" * level, text)


def clean_docstring(text):
    indent = 0
    lines = text.splitlines()
    if len(lines) > 1:
        index = 0
        while index < len(lines) and lines[index].replace(" ", "").replace("\t", ""):
            index += 1
            continue
        index -= 1
        indent = len(lines[index]) - len(lines[index].lstrip())

    newlines = [lines[0]]
    for line in lines[1:]:
        newlines.append(line[indent:].rstrip())
    return "\n".join(newlines)


def render_ref_link(ref):
    target = get_ref_name(ref)
    return "[{}](#{})".format(target, target.lower())


def render_ref_link_html(ref):
    target = get_ref_name(ref)
    return '<a href="#{}">{}</a>'.format(target.lower(), target)


def render_property_default(propertydata):
    default = propertydata.get("default", "")
    return default


def render_property_type(propertydata):
    propertytype = propertydata.get("type", "string")
    if propertytype == "array":
        if "type" in propertydata["items"]:
            reftype = propertydata["items"]["type"]
        elif "$ref" in propertydata["items"]:
            reftype = render_ref_link_html(propertydata["items"]["$ref"])
        else:
            raise NotImplementedError("no item type defined")
        return reftype + " " + propertytype
    elif propertydata.get("anyOf"):
        anyof = propertydata.get("anyOf")
        reftype = next(of["type"] for of in anyof if "type" in of and "$ref" not in of)
        refref = next(of["$ref"] for of in anyof if "$ref" in of and "type" not in of)
        reflink = render_ref_link_html(refref)
        if reftype == "null":
            return reflink
        else:
            raise NotImplementedError("not sure how to handle anyOf")
    return propertytype


TABLE_TEMPLATE = """
<table>
  <thead>
    <tr>
      <th>Field</th>
      <th>Default</th>
      <th>Description</th>
    </tr>
  </thead>
  <tbody>
    {% for property in properties -%}
    <tr>
      <td><strong><code>{{ property.field }}</code></strong>{% if property.required -%}*{% endif -%}
        {% if property.type -%}<br/><em>{{ property.type }}</em>{% endif -%}
        {% if property.deprecated -%}<br/><em>(deprecated)</em>{% endif -%}
      </td>
      <td>
        {% if property.default -%}<code>{{ property.default }}</code>{% endif -%}
      </td>
      <td>
        {{ property.description | markdown -}}
        {% for example in property.examples -%}
        <pre><code>{{ example }}</code></pre>
        {% endfor -%}
      </td>
    </tr>
    {% endfor -%}
  </tbody>
</table>
"""


def render_property_table(struct):
    env = Environment()
    env.filters["markdown"] = to_markdown
    template = env.from_string(TABLE_TEMPLATE)
    properties = []
    for propertyname, propertydata in struct["properties"].items():
        default = render_property_default(propertydata)
        properties.append(
            {
                "field": propertyname,
                "type": render_property_type(propertydata),
                "default": str(default) if str(default) else "",
                "description": propertydata.get("description", ""),
                "examples": propertydata.get("examples", ""),
                "required": not str(default),
                "deprecated": propertydata.get("deprecated", False),
            }
        )

    return template.render(properties=properties)


def render_struct_blocks(struct, header_level):
    blocks = []
    if "description" in struct:
        blocks.append(clean_docstring(struct["description"]))
    blocks.append(render_property_table(struct))
    return blocks


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("input_file", type=argparse.FileType("r"))
    parser.add_argument(
        "-o", "--output-file", type=argparse.FileType("w"), default=sys.stdout
    )
    parser.add_argument("-p", "--package-path")
    parser.add_argument("-l", "--header-level", type=int, metavar="NUM", default=1)
    parser.add_argument("-t", "--title")
    parser.add_argument("-d", "--description")
    args = parser.parse_args()

    schema = json.load(args.input_file)

    blocks = []

    if args.title:
        blocks.append(render_header(args.header_level, args.title))

    if args.description:
        blocks.append(args.description)

    first_name = get_ref_name(schema["$ref"])
    first_struct = schema["$defs"].pop(first_name)

    blocks.extend(
        [
            render_header(args.header_level + 1, first_name),
            *render_struct_blocks(first_struct, args.header_level),
        ]
    )

    for structname, struct in sorted(schema["$defs"].items()):
        blocks.extend(
            [
                render_header(args.header_level + 1, struct["title"]),
                *render_struct_blocks(struct, args.header_level + 1),
            ]
        )

    args.output_file.write("\n\n".join(blocks) + "\n")


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Union

from ruamel.yaml.scalarstring import (
    DoubleQuotedScalarString,
    FoldedScalarString,
    PreservedScalarString,
)


def merge_dict(orig: dict, new: dict) -> dict:
    final = {}

    all_keys = sorted(list(set(orig.keys()) | set(new.keys())))
    for key in all_keys:
        if key in orig and key in new:
            if isinstance(orig[key], dict):
                if not isinstance(new[key], dict):
                    raise NotImplementedError("Can't merge dict and non-dict keys")
                final[key] = merge_dict(orig[key], new[key])
            else:
                final[key] = new[key]  # new takes precedence
        elif key in orig:
            final[key] = orig[key]
        elif key in new:
            final[key] = new[key]
        else:
            raise NotImplementedError("This should not possible")
    return final


def make_multiline_string(line):
    return PreservedScalarString(line)


def make_scalar_string(line, quote=False):
    if "\n" in line.strip():
        return PreservedScalarString(line)
    elif len(line) >= 80:
        return FoldedScalarString(line)
    elif quote:
        return DoubleQuotedScalarString(line)
    return line


def make_quoted_string(line):
    return make_scalar_string(line, quote=True)


def make_scalar_list(value: list[str]) -> list[Any]:
    return [make_scalar_string(line) for line in value]


def make_quoted_list(value: list[str]) -> list[Any]:
    return [make_scalar_string(line, quote=True) for line in value]


def make_quoted_list_or_string(object: Union[str, list[str]]) -> list[Any]:
    if isinstance(object, str):
        return make_quoted_string(object)
    return make_quoted_list(object)


def make_quoted_dict(object):
    return {key: make_quoted_string(value) for key, value in object.items()}



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import argparse
import json
import sys

import msgspec
from ruamel.yaml import YAML

from ..config.project import models as cici_config


def get_schema_type(schema_name: str) -> type[cici_config.File]:
    return {
        "cici-config": cici_config.File,
    }[schema_name]


def write_schema_file(schema, output_file, output_format):
    if output_format == "json":
        json.dump(schema, output_file, indent=2)
    elif output_format == "yaml":
        yaml = YAML(typ="safe")
        yaml.default_flow_style = False
        yaml.dump(schema, output_file)
    else:
        raise NotImplementedError("this is not a supported output format")


def schema_command(parser, args):
    schema_type = get_schema_type(args.schema_name)
    schema = msgspec.json.schema(schema_type)
    write_schema_file(
        schema=schema, output_file=args.output_file, output_format=args.output_format
    )


def schema_parser(subparsers):
    parser = subparsers.add_parser("schema", help="generate json schema")
    parser.add_argument(
        "-s",
        "--schema",
        dest="schema_name",
        choices=["cici-config"],
        default="cici-config",
    )
    parser.add_argument(
        "-f",
        "--format",
        dest="output_format",
        choices=["json", "yaml"],
        default="json",
    )
    parser.add_argument(
        "-o",
        "--output",
        dest="output_file",
        type=argparse.FileType("w"),
        default=sys.stdout,
    )
    parser.set_defaults(func=schema_command)
    return parser



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0


import logging
import re
from importlib import import_module
from io import StringIO
from pathlib import Path
from typing import Optional

from msgspec.structs import replace
from termcolor import colored

from cici.config.project import models as cici_config
from cici.config.project import serializers as cici_config_serializers

from ..constants import DEFAULT_PROVIDER
from ..paths import get_cici_config_path


def get_bundle_name(text):
    return re.sub(r"[-_ ]+", "-", text.casefold())


def get_bundle_content(provider, bundle, cici_config_file=None):
    content = StringIO()
    content.write(f"#\n")
    content.write(
        f"# This file is generated by cici-tools with the following command:\n"
    )
    content.write(f"#\n")
    content.write(f"#     cici bundle\n")
    content.write(f"#\n")
    provider.dump(
        bundle, content, cici_config_file=cici_config_file
    )  # where YAML is written
    return content.getvalue()


def load_cici_or_nothing(base_path: Path) -> Optional[cici_config.File]:
    # try to load and resolve .cici/config.yaml
    # - if config.yaml is there, load it, if not load nothing
    # - if config.yaml does not exist return None

    base_path = Path(base_path).resolve()

    if base_path.name == ".cici":
        base_path = base_path.parent

    config_path_root = base_path / ".cici" / "config.yaml"

    # load config.yaml if it exists
    config_path = None
    if config_path_root.exists():
        config_path = config_path_root

    if config_path:
        config_file = cici_config_serializers.load(config_path)
        # if load() succeeds then resove targets
        return cici_config_serializers.resolve_targets(config_file, config_path)

    return None


def bundle_command(parser, args):

    logging.basicConfig(level=logging.INFO, format="[%(levelname)s] %(message)s")

    provider = import_module(f".{DEFAULT_PROVIDER}", "cici.providers")

    cici_config_file = load_cici_or_nothing(args.config_path)

    if not cici_config_file:
        logging.warning(f"No config.yaml found in .cici")

    ci_file_path = args.config_path / provider.CI_FILE

    if not Path(ci_file_path).exists():
        parser.error(f"file not found: {ci_file_path}")

    file = provider.load(ci_file_path, cici_config_file=cici_config_file)

    args.output_path = Path(args.output_path)

    for job_name, job in file.jobs.items():
        if job_name.startswith("."):
            continue

        # replace deepcopy + direct mutation with msgspec replace
        bundle = replace(file, jobs={job_name: job})

        bundle_filename = args.output_path / f"{get_bundle_name(job_name)}.yml"
        content = get_bundle_content(
            provider, bundle, cici_config_file=cici_config_file
        )

        # ensure consistent whitespace
        content = "\n".join([line.rstrip() for line in content.splitlines()]) + "\n"

        if bundle_filename.exists() and bundle_filename.is_file():
            old_content = open(bundle_filename).read()
            if old_content == content:
                print("  ", colored(bundle_filename.name, "grey"))
                continue

        with open(bundle_filename, "w") as stream:
            stream.write(content)
        print(colored("⚡", "yellow"), bundle_filename.name)


def bundle_parser(subparsers):
    parser = subparsers.add_parser(
        "bundle", help="bundle CI jobs with dependencies into standalone distributions"
    )
    parser.add_argument(
        "config_path",
        metavar="DIR",
        nargs="?",
        type=Path,
        default=get_cici_config_path(),
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="DIR",
        dest="output_path",
        type=Path,
        default=Path.cwd().absolute(),
    )
    parser.set_defaults(func=bundle_command)
    return parser



# SPDX-FileCopyrightText: UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import importlib

import pytest


@pytest.mark.parametrize(
    "module_name",
    [
        "cici",
        "cici._version",
        "cici.__main__",
        "cici.cli",
        "cici.cli.bundle",
        "cici.cli.readme",
        "cici.cli.schema",
        "cici.cli.update",
        "cici.config",
        "cici.config.project",
        "cici.config.project.models",
        "cici.config.project.serializers",
        "cici.config.user",
        "cici.constants",
        "cici.main",
        "cici.providers.gitlab.serializers",
        "cici.providers.gitlab.constants",
        "cici.providers.gitlab.utils",
        "cici.providers.gitlab.models",
        "cici.providers.gitlab",
        "cici.providers",
        "cici.schema",
        "cici.templates",
        "cici.utils",
    ],
)
def test_import_module(module_name):
    importlib.import_module(module_name)
